#!/bin/sh

# Change Log - v1.5
# Eliminated some redundancy within configure vib
# Configure Vib now creates a proper config string using GPU#'s
# Added prompts for simplified configuration with Radeon Pro Settings 

# Change Log - v1.3
# Allows 16 VFs to be configured per GPU when running on ESXi 6.5.0
# Fixed a bug where adding more GPU's to the host after initial configuration would cause a crash

# Change Log - v1.2
# Allows mixed configurations where some GPUs are in passthrough mode while others are virtualized.
# Datastore/directory option in vms.cfg. Consult MxGPU guide for help with configuring vms.cfg
# PCI Hole settings will only be added to the VMs specified by vms.cfg instead of all VMs.
# Fixed an issue where PCI Hole settings were added to VMs that no longer existed.
# Improved interface of the prompts
# Predictable ordering of which VFs will be assigned. All VFs from each GPU will be assigned to VMs before moving on to the VFs on the next GPU.

stty erase ^?

max_VFnum(){
    mainv=$(vmware -v | cut -d " " -f 3 | cut -d "." -f 1)
    subv=$(vmware -v | cut -d " " -f 3 | cut -d "." -f 2)
    
    if [ $mainv -gt 6 ]; then
        max_vfnum=16
    elif [ $mainv -eq 6 ] && [ $subv -ge 5 ]; then
        max_vfnum=16
    else
        max_vfnum=15
    fi
}

cardVIDI="1002:6929"

default_VFnum=4
default_fbsize=1920
default_ts=5000

max_VFnum
min_vfnum=1
total_fbsize=7680
max_fbsize=3952
min_fbsize=256
max_ts=500000
min_ts=3000
predictable_perf="Y"

longret=""
tonga_num=0

getArrayVal(){
    target=${1}${2}
    eval "longret=$`echo ${target}`"
}

setArrayVal(){
    target=${1}${2}
    val=${3}
    eval "`echo ${target}`=${val}"
}


charHexToDec(){

    case $1 in
    "f" )
    ret="15"
    ;;
    "e" )
    ret="14"
    ;;
    "d" )
    ret="13"
    ;;
    "c" )
    ret="12"
    ;;
    "b" )
    ret="11"
    ;;
    "a" )
    ret="10"
    ;;
    * )
    ret=$1
    ;;
    esac

    return $ret
}

hexToDec(){

    htd_out="0"
    htd_len=${#1}

    for htd_i in $(seq $htd_len)
    do
        htd_num=`echo $1 | cut -c $htd_i`
        charHexToDec $htd_num
        htd_out=`expr $htd_out \* 16 + $?`
    done
    longret=$htd_out

}

floor16M(){
    input_val=$1
    remainder=`expr $input_val % 16 `
    if [ $remainder -eq 0 ]
    then
        longret=$input_val
    else
        input_val=`expr $input_val - $remainder`
        longret=$input_val
    fi
    return

}

findVib(){
    if [ ${#1} -eq 0 ] ; then
        num_vibs=$(find . -name "amdgpuv*.vib" | wc -l)
        if [ $num_vibs -eq 1 ]; then
            vibfile=$(find . -name "amdgpuv*.vib")
            echo "amdgpuv driver $vibfile will be installed."
        else
            echo "Enter the name of the amdgpuv driver that you want to install."
            usage
        fi
    else
        vibfile=$1
    fi
} 

installVib(){

    #Check if already installed

    result=`esxcli software vib get --vibname amdgpuv | grep -c "NoMatchError"`
    if [ $result -eq "0" ]
    then
        ret=0
        echo "Driver has been already installed"
        return $ret
    fi

    #Start installation
    vim-cmd hostsvc/maintenance_mode_enter > /dev/null

    echo "+++++Installing driver"
    result=`esxcli software vib install --no-sig-check -v "$(pwd)/$1"`

    vim-cmd hostsvc/maintenance_mode_exit > /dev/null

    ifsucceeded=`echo $result|grep -c "Operation finished successfully"`
    if [ $ifsucceeded -eq "1" ]
    then
        ret=0
        echo "+++++Installation Completed Successfully!"
        esxcli system module set -m amdgpuv -e true
    else
        ret=1
        echo "+++++Installation Failed!"
        echo $result
    fi

    return $ret
}

uninstallVib(){

    result=`esxcli software vib get --vibname amdgpuv-cim | grep -c "NoMatchError"`
    if [ $result -eq "0" ]
    then
        uninstallCIM
    fi
    # Check if already uninstalled
    echo "+++++Uninstalling driver"
    result=`esxcli software vib get --vibname=amdgpuv | grep -c "NoMatchError"`
    if [ $result -eq "1" ]
    then
        echo "No amdgpuv driver installed"
        ret=2
        return $ret
    fi
    vim-cmd hostsvc/maintenance_mode_enter > /dev/null
    
    result=`esxcli software vib remove --vibname=amdgpuv`
    
    vim-cmd hostsvc/maintenance_mode_exit > /dev/null
    
    ifsucceeded=`echo $result|grep -c "Operation finished successfully"`
    if [ $ifsucceeded -eq "1" ]
    then
        echo "+++++amdgpuv Driver Successfully Removed!"
        echo "The configuration needs a reboot to take effect"
        ret=0
    else
        echo "+++++amdgpuv Removal Failed!"
        echo $result
        ret=1
    fi
    return $ret
}

checkVib() {
    result=`esxcli software vib get --vibname amdgpuv | grep -c "NoMatchError"`
    if [ $result -eq "1" ]
    then
        echo "Driver not installed!"
        echo "Run with -i flag to install VIB driver."
        exit
    fi
}

installCIM() {
    result=`esxcli software vib get --vibname amdgpuv-cim | grep -c "NoMatchError"`
    if [ $result -eq "0" ]
    then
        ret=0
        echo "CIM has been already installed"
        return $ret
    fi
    
    if [ ${#1} -eq 0 ] ; then
        num_vibs=$(find . -name "amdgpuv-cim*.vib" | wc -l)
        if [ $num_vibs -eq 1 ]; then
            cimfile=$(find . -name "amdgpuv-cim*.vib")
            echo "amdgpuv-cim $cimfile will be installed."
        else
            echo "Run the script again with the --cim option and specify the name of the amdgpuv-cim vib file"
            exit
        fi
    else
        cimfile=$1
    fi
    
    #Start installation
    vim-cmd hostsvc/maintenance_mode_enter > /dev/null

    echo "+++++Installing CIM"
    result=`esxcli software vib install --no-sig-check -v "$(pwd)/$cimfile"`

    vim-cmd hostsvc/maintenance_mode_exit > /dev/null

    ifsucceeded=`echo $result|grep -c "Operation finished successfully"`
    if [ $ifsucceeded -eq "1" ]
    then
        ret=0
        echo "+++++Installation Completed Successfully!"
        
    else
        ret=1
        echo "+++++Installation Failed!"
        echo $result
    fi
    
    return $ret
}

uninstallCIM() {

    echo "+++++Uninstalling CIM"
    result=`esxcli software vib get --vibname=amdgpuv-cim | grep -c "NoMatchError"`
    if [ $result -eq "1" ]
    then
        echo "No CIM installed"
        ret=2
        return $ret
    fi
    
    vim-cmd hostsvc/maintenance_mode_enter > /dev/null
    
    result=`esxcli software vib remove --vibname=amdgpuv-cim`
    
    vim-cmd hostsvc/maintenance_mode_exit > /dev/null
    
    ifsucceeded=`echo $result|grep -c "Operation finished successfully"`
    if [ $ifsucceeded -eq "1" ]
    then
        echo "+++++CIM Successfully Removed!"
        echo "The configuration needs a reboot to take effect"
        ret=0
    else
        echo "+++++CIM Removal Failed!"
        echo $result
        ret=1
    fi
    return $ret
}

detectCard(){

    #Detect Tonga
    tonga_num=`lspci -v | grep -n $1 | sed -n "$="`
    if [ ${#tonga_num} -eq 0 ]
    then
        tonga_num=0
    fi
    echo "Detected "$tonga_num" SR-IOV GPU"
    
    for i in $(seq $tonga_num)
    do
        line=`lspci -v | grep -n $1 | cut -d : -f 1  | sed -n $i"p"`
        line=`expr $line - 1`
        pci_info=`lspci -v | sed -n $line"p"`
        echo $pci_info
        setArrayVal "pci_buses" $i `echo $pci_info | cut -d : -f 2`
        getArrayVal "pci_buses" $i
    done
}

getInput(){
    default_value=$2
    input_val=""
    read -p "$1" input_val
    if [ ${#input_val} -eq "0" ]
    then
        longret=$2
    else
        longret=$input_val
    fi
    return
}

checkValue(){
    max_val=$2
    min_val=$3
    input_val=$1
    if [ $input_val -lt $min_val ]
    then
        echo "Input value $input_val is smaller than the minimum value $min_val! Set it to the minimum value."
        longret=$min_val
    elif [ $input_val -gt $max_val ]
    then
        echo "Input value $input_val is bigger than the maximum value $max_val! Set it to the maximum value."
        longret=$max_val
    else
        longret=$input_val
    fi
    return
}

checkFBsize(){
    vfnum=$1
    fbsize=$2
    temp=`expr $vfnum \* $fbsize`
    if [ $temp -le $total_fbsize ]
    then
        floor16M $fbsize
    else
        echo "Frame buffer size is too large for this VF num. Auto adjust it to a proper value."
        temp=`expr $total_fbsize / $vfnum`
        floor16M $temp
    fi
    return
}

getConfig(){
    getInput "Please enter number of VFs:(default:$default_VFnum): " $default_VFnum
    checkValue $longret $max_vfnum $min_vfnum
    numVFs=$longret


    temp=`expr $total_fbsize / $numVFs`
    floor16M $temp
    default_fbsize=$temp

    getInput "Please enter FrameBuffer Size(in MB)(default:$default_fbsize): " $default_fbsize
    checkValue $longret $max_fbsize $min_fbsize
    checkFBsize $numVFs $longret
    fbsize=$longret

    if [ $numVFs -eq 2 ]; then
        default_ts=15000
    elif [ $numVFs -eq 4 ]; then
        default_ts=5000
    else
        default_ts=6000
    fi

    getInput "Please enter time slice(in us)(default:$default_ts): " $default_ts
    checkValue $longret $max_ts $min_ts
    ts=$longret
    
    #read -p "Do you want to enable Predictable Performance? ([Y]es/[N]o,default:N): " predictable_perf
}

getConfigAuto(){
    getInput "Please enter number of VFs:(default:$default_VFnum): " $default_VFnum
    checkValue $longret $max_vfnum $min_vfnum
    numVFs=$longret

    temp=`expr $total_fbsize / $numVFs`
    floor16M $temp
    fbsize=$temp

    if [ $numVFs -eq 2 ]; then
        default_ts=15000
    elif [ $numVFs -eq 4 ]; then
        default_ts=5000
    else
        default_ts=6000
    fi
    ts=$default_ts
    
    #read -p "Do you want to enable Predictable Performance? ([Y]es/[N]o,default:N): " predictable_perf
}

checkCompatibility(){
    #Check if this system support VT/AMD-V
    echo "Checking environment......"

    #Check VT/AMD-V capability
    vcap=`esxcfg-info | grep "\----\HV Support\." | sed "s/.*\.//g"`
    
    case $vcap in
        "0" | "1" ) 
            # 0 indicates that VT/AMD-V is not available for this hardware
            # 1 indicates that VT/AMD-V might be available but not supported for this hardware
            echo "VT or AMD-V is not available on this platform"
            exit 0
            ;;
        "2" )
            echo "VT or AMD-V is available on this platform but not enabled in the BIOS"
            exit 0
            ;;
        "3" | * )
            #echo "VT or AMD-V is enabled"
            ;;
    esac    
    
    # Check if Scripts folder is placed in the correct directory
    if [ ! -d "`pwd`/scripts" ]; then
        echo "SCRIPTS FOLDER NOT FOUND!"
        echo "Please copy scripts folder to `pwd` to continue"
        exit
    fi

    # Check memory range
    # Check to make sure file exists before continuing
    if [ ! -f "scripts/mem_range.py" ]; then
        echo "mem_range.py not found!"
        echo "Please re-download \"MxGPU Installation Utility\" from amd.com"
        exit
    fi
    res=`python scripts/mem_range.py`
    if [ $res == "n" ]
    then
        echo "Memory range is not within 40bit addressing."
        exit 0
    fi
}

addPciHole(){
    # Check if Python Script exists
    if [ ! -f "scripts/pcihole.py" ]; then
        echo "pcihole.py not found!"
        echo "Please re-download \"MxGPU Installation Utility\" from amd.com"
        exit
    fi  
    
    #Write pciHole to vmxs
    python scripts/pcihole.py add
}

removePciHole(){
    # Check if Python Script exists
    if [ ! -f "scripts/pcihole.py" ]; then
        echo "pcihole.py not found!"
        echo "Please re-download \"MxGPU Installation Utility\" from amd.com"
        exit
    fi  
    
    #Remove pciHole from vmxs
    python scripts/pcihole.py remove
    
}

#remove any GPU's configured for passthrough
removePassthru() {
    
    numPT=`lspci -p |grep 1002:6929 | grep Passthru -c`
    i=1 
    
    cp /etc/vmware/esx.conf /etc/vmware/esx-"`date`".conf
    while [ $i -le $numPT ];
    do
        bus=`lspci -p | grep $cardVIDI | grep Passthru | cut -d" " -f 1 | sed -n "$i"'p'`
        hex=`echo $bus | cut -d ":" -f 2`
        hexToDec $hex
        hexBus=${bus//$hex/$longret}
        
        device="$hexBus/owner"
        
        #find the passthrough GPU in esx.conf and set it to non-passthrough by changing the owner from passthru to vmkernel
        sed -i '\%[$device]%s%passthru%vmkernel%g' /etc/vmware/esx.conf
        i=$((i+1))
    done
    
    echo "All Passthru devices have been unconfigured."
    
}

configureVib(){

    detectCard $cardVIDI
    if [ $tonga_num -eq 0 ]
    then
        echo "No SRIOV GPU found."
        exit 0
    fi
    
    #Check if there are any GPU's set to pass-through
    numPT=`lspci -p |grep $cardVIDI | grep Passthru -c`
        
    if [ "$numPT" -gt 0 ]; then
        echo "$numPT GPU's are configured for Passthrough. These devices need to be unconfigured for MxGPU."
        read -p "Do you want to unconfigure all AMD devices set to Passthrough? ([Y]es/[N]o, default:Y)" res
        case "$res" in
        [Nn]* )
            echo "No settings were changed! "
            echo "Passthrough devices will not be virtualized."
            
            ;;
        * )
            removePassthru
            ;;
        esac
    fi
    echo "Start configuration...."
    config_str=""
    flag=0;
    
    read -p "Do you plan to use the Radeon Pro Settings vSphere plugin to configure MxGPU? ([Y]es/[N]o, default:N)" web
    
    # check MxGPUi is enabled...
    case "$web" in 
    [Yy]* ) 
        echo "Configured for use with vSphere Plugin"
        if [[ ! -z "`vmkload_mod -s amdgpuv | grep MxGPUi`" ]]; then
            echo "Note: All $max_vfnum virtual functions will be enabled on the bus."
            config_str="MxGPUi=1"
        fi
        numVFs=4
        flag=2
        ;;
    * )     
        echo "Default Mode"
        if [[ ! -z "`vmkload_mod -s amdgpuv | grep MxGPUi`" ]]; then
            config_str="MxGPUi=0"
        fi
        read -p "Enter the configuration mode([A]uto/[H]ybrid,default:A)" mode
        # config_str initialize to "MxGPUi={mode}"
        case $mode in  
            [Hh]* )
                echo "Hybrid Mode Selected"
                
                ;;
            [Aa]* | * )
                echo "Auto Mode Selected "
                getConfigAuto;
                flag=1
                ;;
        esac
        ;;
    esac

    i=1
    for i in $(seq $tonga_num)
    do
        echo "Configuring the GPU $i ..."
        getArrayVal "pci_buses" $i
        pci_bus=$longret
        lspci | grep AMD | grep $pci_bus":00.0"
        hexToDec $pci_bus
        thebus=$longret     
        if [ $flag -eq 0 ]; then getConfigAuto; fi ;
        vfstr="GPU${i}=$numVFs,B$thebus"
        echo $vfstr
        config_str=$config_str" "$vfstr
                # [Yy]* )echo "$vfstr,1";config_str="$config_str,1";;
                # * )echo "$vfstr,0";config_str="$config_str,0";;
            # esac          
    done
    echo "Setting up SR-IOV settings..."
    esxcfg-module -s "$config_str" amdgpuv
    echo "Done"
    
    if [ $flag -eq 2 ]; then
        installCIM
    else  
        addPciHole
    fi
        
    echo "The configuration needs a reboot to take effect"
    echo ""
    echo "To automatically assign VFs, please run \"sh mxgpu-install.sh -a\" after system reboot"
    
    
}

installVibSetup(){

    installVib $vibfile
    if [ $? -ne 0 ]
    then
        exit 0
    fi
    
    configureVib
}

multiAssign(){
    vfVIDI="1002:692f"
    # Check to make sure there is the expected number of VFs before assigning
    detectedVFs=`lspci -p | grep $vfVIDI | grep VF -c`
    
    if [ $detectedVFs -gt 0 ]; then
        python scripts/frontstage.py
        flag=true
        while [ $flag == true ];
        do 
            read -p "Please select an option [assign|unassign|list|help]: " opt
            case "$opt" in
            "assign"  | "unassign" )
                python scripts/frontstage.py $opt
                flag=false
                ;;
            "list" | "help" )
                python scripts/frontstage.py $opt
                ;;
            * )
                ;;
            esac
        done
        exit
    fi
    echo "No Virtual Functions detected. Ensure that GPU's that you want to virtualize are not in Passthrough mode."
    #echo "Detected: $detectedVFs VFs, Configured: $configuredVFs VFs"
    echo "Please reboot the system before running \"sh mxgpu-install.sh -a\""
    
}

usage(){
    printf  "
Usage: sh mxgpu-install.sh [-acr] [-i VIBFILE] [-u VIBFILE]
Install, configure, update, auto-assign, or remove AMD amdgpuv driver
\t -i <VIB name>*\t\tInstall and Configure amdgpuv driver
\t -c\t\t\tConfigure amdgpuv driver
\t -a\t\t\tAutomatically assign or unassign VFs to VMs
\t -u <VIB name>*\t\tUpdate to specified amdgpuv driver
\t -r\t\t\tRemove amdgpuv driver and pciHole parameters
\t -ap\t\t\tAdd pciHole parameters
\t -rp\t\t\tRemove pciHole parameters 
\t --cim <VIB name>*\tInstall amdgpuv-cim driver 
*: Optional VIB can be specified, otherwise it will look for an amdgpuv VIB in the current directory.\n";
    exit
}

case "$1" in
    "-r" | "--remove" )
        uninstallVib
        removePciHole
        ;;
    "-i" | "--install" )
        findVib $2
        checkCompatibility
        installVibSetup
        ;;
    "-u" | "--update" | "--upgrade" )
        findVib $2
        uninstallVib
        case "$ret" in
        "1" )
            exit
            ;;
        "2" | * )
            echo "Proceeding with VIB install..."
            ;; 
        esac
        checkCompatibility
        installVibSetup
        ;;
    "-c" | "--configure" )
        checkVib
        configureVib
        ;;
    "-a" | "--assign" )
        checkVib
        multiAssign
        ;;
    "-ap" | "--addpci" )
        addPciHole
        ;;
    "-rp" | "--removepci" )
        removePciHole
        ;;
    "--cim" )
        installCIM $2
        ;;    
    * )
        echo "\"$1\" is not a valid argument."
        echo 
        usage 
        ;;
esac
