import fcntl
from ctypes import *
import types


class amdgpuv_vf_info_t(Structure):
    _pack_ = 1
    _fields_ = [("identifier", c_char * 64), \
                ("funcID", c_int), \
                ("vfBDF", c_char * 8), \
                ("vfFBSize", c_ulonglong), \
                ("vfStatus", c_uint)]
    pass


class amdgpuv_device_info_t(Structure):
    _pack_ = 1
    _fields_ = [("identifier", c_char * 64), \
                ("deviceID", c_uint), \
                ("pfBDF", c_char * 8), \
                ("pfFBSize", c_ushort), \
                ("vfCount", c_uint), \
                ("schedIntv", c_ulonglong), \
                ("vfInfo", amdgpuv_vf_info_t * 16)]


'''
Probes for a device located at the argument location
Args:
    location: String where the device is located at. Usually in /dev/
Returns: 
    amdgpuv_device_info_t: Struct of the device's information 
    null: Null on failure
'''


def probeDevice(location):
    try:
        fd = open(location, 'r+')
    except IOError:
        return None;
    else:
        deviceInfo = amdgpuv_device_info_t()
        fcntl.ioctl(fd, 7, deviceInfo)
        return deviceInfo


'''
Gets all PCI device's located in /dev/
Args:
    None
Returns: 
    dict: A dictionary of all valid virtual functions. Keys are the function BDF,
    values are the assigned VM, initialized to none
'''


def getCardInfo():
    vfBDFs = {}
    # probe locations in dev
    for i in range(32, 40):
        location = '/dev/vmgfx' + str(i)
        # print('location: '+location)
        deviceInfo = probeDevice(location);
        if (deviceInfo == None):
            continue
        # show the device
        # print(deviceInfo.identifier)
        # print('%s,%d,%d,%Ld'%(deviceInfo.pfBDF,deviceInfo.pfFBSize,deviceInfo.vfCount,deviceInfo.schedIntv))
        # devices[deviceInfo.identifier]=deviceInfo
        for j in range(deviceInfo.vfCount):
            temp = str(deviceInfo.vfInfo[j].vfBDF.decode('utf-8')).split(":")
            vfBDFs[(int("0x" + temp[0], 16), 
                    int(temp[1].split(".")[0]), 
                    int(temp[1].split(".")[1])
                    )] = None
            # print(deviceInfo.vfInfo[j].vfBDF)
            # print('-------------------------------------')
    return vfBDFs


'''
    fd=open('/dev/vmgfx32','r+')
    deviceinfo=amdgpuv_device_info_t()
    fcntl.ioctl(fd,7,deviceinfo)
    print(deviceinfo.identifier)
    print('%s,%d,%d,%Ld'%(deviceinfo.pfBDF,deviceinfo.pfFBSize,deviceinfo.vfCount,deviceinfo.schedIntv))
    for vf in deviceinfo.vfInfo:
        print('%s'%(vf.identifier))

'''
