import sys
import re
import fcntl
from systeminfo import *
from tools import *
from cardinfo import *


class SimpleTable:
    def __init__(self):
        self.headers=[]
        self.linedata=[]
        pass
    #header (header_name,maxlen)
    def InitHeader(self,headers):
        self.headers = headers;
    def AddLine(self,data):
        self.linedata.append(data)
    def WholeTable(self):
        #headers
        ret = "|"
        for header in self.headers:
            ret += " " + header[0]
            if len(header[0]) < header[1] - 2:
                ret += " " * (header[1] - 2 - len(header[0]))
            ret += " |"
        ret += "\n"
        #content
        for line in self.linedata:
            ret += "| "
            for count in range(0,len(line)):
                cell = line[count]
                header = self.headers[count]
                if len(cell) > header[1] - 2:
                    cell = cell[0:header[1]-5] + "..."
                elif len(cell) < header[1] - 2:
                    cell = cell + " " * (header[1] - 2 - len(cell))
                ret += cell + " | "
            ret += "\n"
        return ret

def RegMatch(name,target):
    name= name + "\Z"
    if re.match(name,target):
        return True
    else:
        return False
    pass

def FindVMByPath(path, srv):
    ret = []
    for vm in srv.vms:
        if path in vm.vm_vmxpath.replace("\"", ""):
            ret.append(vm)
    return ret

def FindVMRe(name,srv):
    ret = []
    for vm in srv.vms:
        if RegMatch(name,vm.vm_name) == True:
            ret.append(vm)
    return ret

def FindVMBare(name,srv):
    ret = []
    for vm in srv.vms:
        if name == vm.vm_name:
            ret.append(vm)
    return ret

def AssignVM(srv,vmlist):
    for vm in vmlist:
        vfbdf = srv.GetEmptyVF()
        srv.AssignVMByBDF(vm,vfbdf)
    srv.Implement()
    pass

def UnAssignVM(srv,vmlist):
    for vm in vmlist:
        srv.UnAssignVM(vm)
    srv.Implement()
    pass

def OpenDevice(path="/dev/vmgfx32"):
    try:
        return open(path,'r')
    except:
        return None

def GetVMList(options, srv):
    vmlist = []
    find = FindVMByPath
    for name in options["name"]:
        if name == 'Reg:':
            find = FindVMRe
            continue
        elif name == 'NoReg:':
            find = FindVMBare
            continue
        vmlist.extend(find(name,srv))
    #discard same vms
    uniquelist=[]
    for vm in vmlist:
        if vm not in uniquelist:
            uniquelist.append(vm)

    return uniquelist

def DoAssign(params):
    argv=params[0]
    srv = params[2]
    options=params[3]
    summary=srv.GetDevicesSummary()
    vmlist = GetVMList(options, srv)
    unassignedlist = []
    for vm in vmlist:
        if len(vm.GetAMDVFPassthru()) == 0:
            unassignedlist.append(vm)
        else:
            print("VM %s, has already assigned a VF"%(vm.vm_name))
    vmlist = unassignedlist
    if len(vmlist) == 0:
        print("There is no eligible VM, exit")
        sys.exit()
    print("Eligible VMs:")
    for vm in vmlist:
        print(vm.vm_name)
    #if available VF is not enough,then exit
    if len(vmlist) > summary[1]:
        print("There are %d VMs eligible, but there are only %d available VFs left" %(len(vmlist),summary[1]))
        print("Please edit the vms.cfg file to specify %d VMs or increase number of VFs" %(summary[1]))
        sys.exit()
    #if all-yes is not enabled, ask for confirm
    elif "y" not in list(options.keys()):
        question = "These VMs will be assigned a VF"
        yesno=None
        while yesno not in ("y","Y","") and yesno not in ("n","N"):
            yesno=input("%s, is this OK?[Y/N]"%(question))
        if yesno in ("n","N"):
            print("User Exit")
            sys.exit()
    #start assign
    AssignVM(srv,vmlist)
    print("done")

    return 0xffff

def DoUnAssign(params):
    argv=params[0]
    srv = params[2]
    options=params[3]
    summary=srv.GetDevicesSummary()
    vmlist = GetVMList(options, srv)
    assignedlist = []
    #remove the vms which is already unassigned
    for vm in vmlist:
        if len(vm.GetAMDVFPassthru()) > 0:
            assignedlist.append(vm)
        else:
            print("VM %s, has already unassigned"%(vm.vm_name))
    vmlist = assignedlist
    if len(vmlist) == 0:
        print("There is no eligible VM, exit")
        sys.exit()
    print("Eligible VMs:")
    for vm in vmlist:
        print(vm.vm_name)
    #if all-yes is not enabled, ask for confirm
    if "y" not in list(options.keys()):
        question = "VF on these VMs will be unassigned"
        yesno=None
        while yesno not in ("y","Y","") and yesno not in ("n","N"):
            yesno=input("%s, is this OK?[Y/N]"%(question))
        if yesno in ("n","N"):
            print("User Exit")
            sys.exit()
    #start assign
    UnAssignVM(srv,vmlist)
    print("done")

    return 0xffff

def ListVM(params):
    argv=params[0]
    '''if(len(params[0])) <= count + 1:
        print "Missing Parameter: %s" %(argv[count])
        sys.exit()
    name = argv[count + 1]'''
    srv = params[2]
    options=params[3]
    vmlist = FindVMRe(".*",srv)
    #Print table
    table = SimpleTable()
    table.InitHeader([("VM ID",8),("VM NAME",32),("GPU ASSIGNED",16)])
    for vm in vmlist:
        table.AddLine((vm.vm_id,vm.vm_name,"ASSIGNED" if len(vm.GetAMDVFPassthru()) > 0 else "NONE"))
    text = table.WholeTable()
    print(text)
    return 0xffff

def ListCard(params):
    srv = params[2]
    devicetable = SimpleTable()
    devicetable.InitHeader([("CARD NAME",24),("Total VFs",16),("Used VFs",16)])
    #devicepaths = execcmd("find /dev/ -name vmgfx* -maxdepth 1|sort")
    cardcount = 0
    #for path in devicepaths.split('\n'):
        #get pf bus
    devicebuses = ""
    for card in AMD_DEVICE_ID_PF_VF:
        res = execcmd("lspci -n | grep \"1002:%s\" | cut -d ':' -f 2" %(card[0]),srv.netconfig)
        if res != None:
            devicebuses += res
    #if len(devicebuses) == 0:
    for bus in devicebuses.split('\n'):
        #read deviceinfo
        summary = srv.CountDeviceAssignByBus(bus)
        total = summary[0]
        used = summary[1]
        cardcount += 1
        #addline
        devicetable.AddLine((AMD_DEVICE_ID_PF_VF[0][2],"%s"%total,"%s"%used))

    if cardcount > 0:
        print(devicetable.WholeTable())
    else:
        print("No AMD SR-IOV card found")

    return 0xffff

def DoList(params):
    options=params[3]
    if "vm" in list(options.keys()):
        ListVM(params)
    elif "card" in list(options.keys()):
        ListCard(params)
    else:
        ListVM(params)
        ListCard(params)

def DoHelp(params=None):
    help_str="\
    AMD MxGPU Multi-Assign Script\n\
    Usage : [assign|unassign|list|help]\n\
        \n\
    assign : Assign VFs to selected VMs automatically. By default, it will read VM filters from vms.cfg\n\
    unassign : Unassign VFs from the selected VMs. By default, it will read VM filters from vms.cfg\n\
    list : Print vm and card infomation tables\n\
    help : Print this help\n\
        \n\
    Note :\n\
    \tEdit vms.cfg to specify VMs that are eligible for Multi-Assign.\n\
    \tBy default, vms.cfg will detect all registered VMs as eligible.\n\
    "
    print(help_str)

    return 0xffff

def ParseNameFile(name,options):
    fd = open(name,"r")
    options["name"]=[]
    for line in fd.readlines():
        line = line.split("#")[0].strip()
        if len(line) > 0:
            options["name"].append(line)
    fd.close()

#Argument Operation List, return 0xffff means to ignore arguments after it
def ArgvNameFile(params):
    argv=params[0]
    if(len(params[0])) <= count + 1:
        print("Missing Parameter: %s" %(argv[count]))
        sys.exit()
    name = argv[count + 1]
    options=params[3]
    ParseNameFile(name,options)
    return count + 2
def ArgvList(params):
    return count + 1
def ArgvHelp(params):
    return 0xffff
def ArgvAssign(params):
    return count + 1
def ArgvUnAssign(params):
    return count + 1

def ArgvAllYes(params):
    params[3]["y"]=True
    return count + 1

def ArgvName(params):
    argv=params[0]
    if(len(params[0])) <= count + 1:
        print("Missing Parameter: %s" %(argv[count]))
        sys.exit()
    name = argv[count + 1]
    options=params[3]
    options["name"]=[name]
    return count + 2

def ArgvListCard(params):
    params[3]["card"]=True
    return count + 1

def ArgvListVM(params):
    params[3]["vm"]=True
    return count + 1

#main ops
#If main operation has been got, if it is already got, ignore others
main_op_dict=dict()
main_op_dict["assign"]=(ArgvAssign,DoAssign)
main_op_dict["unassign"]=(ArgvAssign,DoUnAssign)
main_op_dict["list"]=(ArgvList,DoList)
main_op_dict["help"]=main_op_dict["-h"]=(ArgvHelp,DoHelp)

#optional ops
opt_op_dict=dict()
opt_op_dict["-y"]=ArgvAllYes
opt_op_dict["--file"]=opt_op_dict["-f"]=ArgvNameFile
opt_op_dict["--vm"]=opt_op_dict["-v"]=ArgvListVM
opt_op_dict["--card"]=opt_op_dict["-c"]=ArgvListCard
opt_op_dict["--name"]=opt_op_dict["-n"]=ArgvName

if __name__=='__main__':
    argv = sys.argv[1:]
    srv = ServerInfo()
    srv.Init()
    if len(srv.cardinfo) == 0:
        print("No capable card found")
        sys.exit()
    count = 0
    main_op = None
    main_op_pos = 0
    options = dict()
    while count < len(argv):
        if argv[count] in list(main_op_dict.keys()) and main_op == None:
            main_op_pos = count
            main_op = main_op_dict[argv[count]][1]
            count = main_op_dict[argv[count]][0]((argv,count,srv,options))

        elif argv[count] in list(main_op_dict.keys()) and main_op != None:
            pass
        elif argv[count] in list(opt_op_dict.keys()):
            count = opt_op_dict[argv[count]]((argv,count,srv,options))
        else:
            print("Unknown option %s"%(argv[count]))
            sys.exit()

    if main_op == None:
        main_op = DoHelp

    if "name" not in list(options.keys()) and main_op in (DoAssign,DoUnAssign):
        ParseNameFile("vms.cfg",options)

    main_op((argv,main_op_pos,srv,options))
