#!/usr/bin/python
# -*- coding: utf-8 -*-
import subprocess
import string
import os
import fcntl
from tools import *
from deviceinfo import *
from cardinfo import *


class CardInfo:

    def __init__(self):
        self.pf = None
        self.vfs = []
        self.device_path = ''
        pass

    def ReadCardInfo(self):
        pass


class PassthruInfo:

    def __init__(self, no=None, parent=None):
        self.parent = parent
        self.present = 'FALSE'
        self.no = no
        self.bdf = ''
        self.device_id = ''
        self.vendor_id = ''
        self.uuid = ''
        self.slot = ''
        self.modified = False
        pass

    # Copy data, only data

    def CopyData(self, newone):

        # no won't be copied, but be inherited

        self.no = self.no
        self.parent = self.parent

        # copy only datas

        self.present = newone.present
        self.bdf = newone.bdf
        self.device_id = newone.device_id
        self.vendor_id = newone.vendor_id
        self.uuid = newone.uuid
        self.slot = newone.slot
        self.modified = True
        pass

    def Enable(self):
        self.modified = True
        self.present = 'TRUE'
        pass

    # Disable this passthru

    def Disable(self):
        self.modified = True
        self.present = 'FALSE'
        pass

    def GenerateConfig(self):
        id_str = 'pciPassthru%s.' % self.no
        bdf = BDF(self.bdf)
        bdf_str = id_str + 'id = "%s"' % (str(bdf[0]) + ':'
                + str(bdf[1]) + '.' + str(bdf[2]))
        deviceid_str = id_str + 'deviceId = "%s"' % self.device_id
        vendorid_str = id_str + 'vendorId = "%s"' % self.vendor_id
        systemid_str = id_str + 'systemId = "%s"' % self.uuid
        present_str = id_str + 'present = "%s"' % self.present

        return (
            id_str,
            bdf_str,
            deviceid_str,
            vendorid_str,
            systemid_str,
            present_str,
            )

    def Print(self):
        if self.present == "TRUE":
            print("\tPassthru %s\n\tBDF=%s\n\tdeviceID=%s\n\tvendorID=%s\n\tsys_uuid=%s\n\tslot=%s\n\tpresent=%s\n" \
            %(self.no,self.bdf,self.device_id,self.vendor_id,self.uuid,self.slot,self.present))
        pass


class VMInfo:

    def __init__(
        self,
        vmid,
        vmxpath,
        name,
        datastore,
        parent,
        ):
        self.vm_id = vmid
        self.vm_vmxpath = vmxpath
        self.vm_name = name
        self.parent = parent
        self.modified = False

        # Init 16 slot for passthrus

        self.passthrus = []
        self.passthru_num = 0
        self.ScanPassthru()
        pass

    # Scan Passthrus on VM

    def ScanPassthru(self):
        scan_rawdata = execcmd('cat ' + self.vm_vmxpath
                               + ' | grep "pciPassthru" | sed "s/pciPassthru//g"'
                               , self.parent.netconfig)
        for line in scan_rawdata.splitlines():
            if len(line) == 0:
                continue
            passthru_no = line.split('.')[0]
            passthru = None
            for ipt in range(self.passthru_num):
                if self.passthrus[ipt].no == passthru_no:
                    passthru = self.passthrus[ipt]
                    break
            if passthru == None:
                passthru = PassthruInfo(passthru_no, self)
                self.passthrus.append(passthru)
                self.passthru_num += 1
                pass

            # cases

            def caseid(param):
                passthru.bdf = param
                pass

            def casevendorid(param):
                passthru.vendor_id = param
                pass

            def casedeviceid(param):
                passthru.device_id = param
                pass

            def casepresent(param):
                passthru.present = param
                pass

            def caseslot(param):
                passthru.slot = param
                pass

            def casesysid(param):
                passthru.uuid = param
                pass

            lineattr = line.split('=')[0].split('.')[1].replace(' ', '')
            linevalue = line.split('=')[1].replace(' ', '').replace('"'
                    , '')
            attrdict = {
                'id': caseid,
                'vendorId': casevendorid,
                'deviceId': casedeviceid,
                'systemId': casesysid,
                'present': casepresent,
                'pciSlotNumber': caseslot,
                }

            for attrkey in attrdict:
                if attrkey == lineattr:
                    attrdict[attrkey](linevalue)
        pass

    def Print(self):
        print("VM %s\nvm_name=%s\nvmxpath=%s" %(self.vm_id,self.vm_name,self.vm_vmxpath))
        for passthru in self.passthrus:
            passthru.Print()

    def GetInfo(self):
        ret = dict()
        ret['vm_id'] = self.vm_id
        ret['vm_name'] = self.vm_name
        tonga = self.GetAMDVFPassthru()
        if len(tonga) > 0:
            ret['card'] = self.parent.GetVFIdentifier(tonga[0].bdf)
        else:
            ret['card'] = ''
        return ret

    # Add a passthru

    def AddPassthru(self, passthru):

        # find if there is a passthru which has been disabled

        for i in self.passthrus:

        # if exists replace it with a new one and enable it

            if i.present == 'FALSE':
                i.CopyData(passthru)
                i.Enable()
                self.modified = True
                return

        # if it doesn't exist, create a new passthru by the No.=num then num+1

        newpassthru = PassthruInfo(self.passthru_num, self)
        newpassthru.CopyData(passthru)
        newpassthru.Enable()
        self.passthrus.append(newpassthru)
        self.passthru_num += 1
        self.modified = True
        pass

    # disable an exist passthru

    def RemovePassthru(self, passthru):
        if passthru == None:
            return
        for i in self.passthrus:
            if i.bdf == passthru.bdf:
                i.Disable()
        self.modified = True
        pass

    def GetAMDVFPassthru(self):
        ret = []
        for passthru in self.passthrus:
            if passthru.vendor_id == '0x1002' and passthru.present \
                == 'TRUE' and passthru.device_id \
                in list(map(DeviceGetVF_Hex, AMD_DEVICE_ID_PF_VF)):
                ret.append(passthru)
        return ret

    def GetModifiedPassthru(self):
        ret = []
        for passthru in self.passthrus:
            if passthru.vendor_id == '0x1002' and passthru.modified \
                == True and passthru.device_id \
                in list(map(DeviceGetVF_Hex, AMD_DEVICE_ID_PF_VF)):
                ret.append(passthru)
        return ret

    def WritePassthruChanges(self):
        for passthru in self.GetModifiedPassthru():
            config_strs = passthru.GenerateConfig()

            # remove the old config, config_strs[0] is the header of config strings

            execcmd('sed -i -e "s/%s.*//g" -e "/^$/d" %s'
                    % (config_strs[0], self.vm_vmxpath),
                    self.parent.netconfig)

            # add new config

            for i in range(1, len(config_strs)):
                execcmd('echo %s >> %s' % (config_strs[i],
                        self.vm_vmxpath), self.parent.netconfig)
        print ('VM ' + self.vm_name + ' configure over')
        execcmd('vim-cmd vmsvc/reload $(vim-cmd vmsvc/getallvms | grep -w "%s" | cut -d " " -f1)'
                 % self.vm_name, self.parent.netconfig)


class ServerInfo:

    def __init__(self, net=None, parent=None):
        self.vm_num = 0
        self.vms = []
        self.netconfig = net
        self.parent = parent
        pass

    # Scan VMs on Server

    def Init(self):
        self.ScanSrvUuid()
        self.ScanVM()
        self.ScanDevices()
        if len(self.cardinfo) > 0:
            self.MatchDevice()

    def ScanVM(self):
        scan_rawdata = execcmd('vim-cmd vmsvc/getallvms | awk "NR>1"',
                               self.netconfig)
        if len(scan_rawdata) == 0:
            return
        for line in scan_rawdata.splitlines():
            if len(line) == 0:
                continue
            res = \
                ReExpMatch('(\d+?)\s{3,}(\S.*?)\s{3,}\[(\S.*)\]\s(\S.*?)\s{3,}'
                           , line)
            if len(res) == 0:
                continue
            vmxpath = '/vmfs/volumes/' + '"' + res[0][2] + '"/"' \
                + res[0][3] + '"'
            self.vms.append(VMInfo(res[0][0], vmxpath, res[0][1],
                            res[0][2], self))
            self.vm_num += 1
        pass

    # Scan the system id

    def ScanSrvUuid(self):
        self.system_uuid = \
            execcmd('vsish -e cat /system/systemUuid | grep "uuid String" | cut -d ":" -f 2'
                    , self.netconfig)
        pass

    def ScanDevices(self):

        # enum devices
        # card info is a dictionary
        # keys are tuples of the vfBDF 
        # values are the assignment 
        self.cardinfo = getCardInfo()
        
        
    # See if a card has already been assigned

    def MatchDevice(self):

        # enum all assigned BDF

        for vm in self.vms:
            tonga = vm.GetAMDVFPassthru()
            if len(tonga) > 0:
                offset = (0 if len(tonga[0].bdf.split(':')) == 2 else 1)
                bus = hex(int(tonga[0].bdf.split(':')[0 + offset]))
                rest = tonga[0].bdf.split(':')[1 + offset]
                tonga = str(bus) + ':' + rest
                if BDF(tonga) in list(self.cardinfo.keys()):
                    self.cardinfo[BDF(tonga)] = vm
                pass
            pass

    def CountDeviceAssignByBus(self, bus):
        total = 0
        used = 0
        for (key, value) in list(self.cardinfo.items()):
            if key[0] == int(bus,16): #int(bus, 16):
                total += 1
                used += (1 if value != None else 0)
        return (total, used)

    def GetDevicesSummary(self):
        total_amount = len(self.cardinfo)
        total_available = list(self.cardinfo.values()).count(None)
        
        return (total_amount, total_available)

    def GetEmptyVF(self):
        for (vfbdf, vm) in sorted(self.cardinfo.items()):
            if vm == None:
                return '%s:%02d.%d' % (hex(vfbdf[0])[2:], vfbdf[1],
                        vfbdf[2])
        return None

    def AssignVMByBDF(self, vm, vfbdf):
        if not vm in self.vms:
            return -1
        if len(vm.GetAMDVFPassthru()) > 0:
            return 0
        passthru = PassthruInfo()
        passthru.bdf = vfbdf
        passthru.device_id = '0x' + execcmd('lspci -n | grep ' + vfbdf
                + ' | cut -d " " -f 4 | cut -d ":" -f 2',
                self.netconfig)
        passthru.vendor_id = '0x1002'
        passthru.uuid = self.system_uuid
        vm.AddPassthru(passthru)
        self.cardinfo[BDF(vfbdf)] = vm

    def UnAssignVM(self, vm):
        if not vm in self.vms:
            return -1
        for passthru in vm.GetAMDVFPassthru():
            vm.RemovePassthru(passthru)
        pass

    def Implement(self):
        for vm in self.vms:
            if vm.modified == True:
                vm.WritePassthruChanges()

    def GetInfo(self):
        ret = []
        for vm in self.vms:
            ret.append(vm.GetInfo())
        return ret

    def Print(self):
        for vm in self.vms:
            vm.Print()


class VCenterInfo:

    def __init__(self, conf='./serverlist'):
        self.serverlist = []
        self.conf = conf
        self.ScanSys()
        pass

    # Scan Systems on vCenter

    def ScanSys(self):
        f = open(self.conf, 'r')
        for line in f.readlines():
            data = line.split()
            if len(data) == 3:
                print ('Checking connection to %s...' % data[0])
                if execcmd('ls', (data[0], data[1], data[2])) == None:
                    print ('Can not connect to %s, Skip...' % data[0])
                else:
                    srv = ServerInfo(net=(data[0], data[1], data[2]))
                    srv.Init()
                    self.serverlist.append(srv)
            else:
                print ('serverlist syntax error!')
                break
        f.close()
        pass

    def GetServerIPList(self):
        ret = []
        for srv in self.serverlist:
            ret.append(srv.ip)
        return ret

    def Print(self):
        for srvobj in self.serverlist:
            print ('Server:' + srvobj.netconfig[0])
            srvobj.Print()
        pass


if __name__ == '__main__':
    srv = ServerInfo()
    srv.Init()
    print (srv.Print())

            